/** MasterGear: portable MasterSystem/GameGear emulator ******/
/**                                                         **/
/**                           Unix.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995,1996            **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef UNIX

/** Private #includes ****************************************/
#include "M6502.h"
#include "fce.h"
#include "palette.h"

#ifdef SOUND
#include "DSP.h"
#endif

/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

/** MIT Shared Memory Extension for X ************************/
#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
XShmSegmentInfo SHMInfo;
int UseSHM=1;
#endif

/** Sound-related definitions ********************************/
#ifdef SOUND
int UseSound  = 0;
int SndSwitch = 0x0F;
int SndVolume = 200;
#endif

/** Various X-related variables ******************************/
Display *Dsp;
Window Wnd;
Colormap DefaultCMap;
XImage *Img;
GC DefaultGC;
unsigned long White,Black;

/** Various variables and short functions ********************/
#define WIDTH  256+16
#define HEIGHT 240

extern        M6502 *R;

char *Title="fce Unix/X ";

int SaveCPU=1;

unsigned short   *XBuf;
unsigned short  XPal[16],SPal[16];

int Verbose= 1;

void OnBreak(int Arg) {  /*  CPURunning=0 */ ; }


void PutImage(int X,int Y,int W,int H)
{
#ifdef MITSHM
  if(UseSHM)
    XShmPutImage
    (Dsp,Wnd,DefaultGC,Img,X,Y,(WIDTH-W)/2,(HEIGHT-H)/2,W,H,False);
  else
#endif
    XPutImage
    (Dsp,Wnd,DefaultGC,Img,X,Y,(WIDTH-W)/2,(HEIGHT-H)/2,W,H);

  XFlush(Dsp);
}

/** SetColor *************************************************/
/** Allocate new color (0-15 picture, 16-31 sprites).       **/
/*************************************************************/
void SetColor(byte N,byte R,byte G,byte B)
{
  unsigned short *P;
  if(N>15) { N-=16;P=SPal; } else P=XPal;
  P[N]=Pal[((R&0xC0)>>2)|((G&0xC0)>>4)|((B&0xC0)>>6)];
}

/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  long L;
  int J;

  if(Verbose) printf("Shutting down...\n");

  if(Dsp&&Wnd)
  {
#ifdef MITSHM
    if(UseSHM)
    {
      XShmDetach(Dsp,&SHMInfo);
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
    }
    else
#endif MITSHM
    if(Img) XDestroyImage(Img);
    for(J=0;J<64;J++)
      if(Pal[J]!=Black) { L=Pal[J];XFreeColors(Dsp,DefaultCMap,&L,1,0); }
    XAutoRepeatOn(Dsp);
  }

  if(Dsp) { XAutoRepeatOn(Dsp);XCloseDisplay(Dsp); }

#ifdef SOUND
  TrashSound();
#endif SOUND
}

/** InitMachine **********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  Screen *Scr;
  XEvent E;
  XColor Color;
  int J,I;

  XBuf=NULL;

  if(Verbose)
    printf("Initializing Unix/X drivers:\n  Opening display...");
  Dsp=XOpenDisplay(NULL);
  if(!Dsp) { if(Verbose) printf("FAILED\n");return(0); }

  Scr=DefaultScreenOfDisplay(Dsp);
  White=WhitePixelOfScreen(Scr);
  Black=BlackPixelOfScreen(Scr);
  DefaultGC=DefaultGCOfScreen(Scr);
  DefaultCMap=DefaultColormapOfScreen(Scr);

  for(J=0;J<64;J++) Pal[J]=Black;

  if(Verbose) printf("OK\n  Opening window...");
  Wnd=
    XCreateSimpleWindow
    (Dsp,RootWindowOfScreen(Scr),0,0,WIDTH,HEIGHT,0,White,Black);
  if(!Wnd) { if(Verbose) printf("FAILED\n");return(0); }

  {
    XSizeHints Hints;
    XWMHints WMHints;

    Hints.flags=PSize|PMinSize|PMaxSize;
    Hints.min_width=Hints.max_width=Hints.base_width=WIDTH;
    Hints.min_height=Hints.max_height=Hints.base_height=HEIGHT;
    WMHints.input=True;WMHints.flags=InputHint;
    XSetWMHints(Dsp,Wnd,&WMHints);
    XSetWMNormalHints(Dsp,Wnd,&Hints);
    XStoreName(Dsp,Wnd,Title);
  }

  XSelectInput
  (Dsp,Wnd,FocusChangeMask|ExposureMask|KeyPressMask|KeyReleaseMask);
  XMapRaised(Dsp,Wnd);
  XClearWindow(Dsp,Wnd);
  XAutoRepeatOff(Dsp);
  XWindowEvent(Dsp,Wnd,ExposureMask,&E);

#ifdef MITSHM
  if(UseSHM)
  {
    if(Verbose) printf("OK\n  Using shared memory:\n    Creating image...");
    Img=
      XShmCreateImage
      (Dsp,DefaultVisualOfScreen(Scr),16,ZPixmap,NULL,&SHMInfo,WIDTH,HEIGHT);
    if(!Img) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Getting SHM info...");
    SHMInfo.shmid=
      shmget(IPC_PRIVATE,Img->bytes_per_line*Img->height,IPC_CREAT|0777);
    if(SHMInfo.shmid<0) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Allocating SHM...");
    XBuf=(unsigned short *)(Img->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    SHMInfo.readOnly=False;
    if(Verbose) printf("OK\n    Attaching SHM...");
    if(!XShmAttach(Dsp,&SHMInfo))
    { if(Verbose) printf("FAILED\n");return(0); }
  }
  else
#endif
  {
    if(Verbose) printf("OK\n  Allocating screen buffer...");
    XBuf=(unsigned short *)malloc(sizeof(unsigned short)*HEIGHT*WIDTH);
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n  Creating image...");
    Img=
      XCreateImage
      (Dsp,DefaultVisualOfScreen(Scr),16,ZPixmap,0,(char *)XBuf,WIDTH,HEIGHT,8,0);
    if(!Img) { if(Verbose) printf("FAILED\n");return(0); }
  }

  if(Verbose) printf("OK\n  Allocating palette...");
  for(J=I=0;J<64;J++)
  {
    Color.flags=DoRed|DoGreen|DoBlue;
    Color.red=	(palette[J].r)<<10;
    Color.green=(palette[J].g)<<10;
    Color.blue=	(palette[J].b)<<10;
#ifdef 0
printf("%d\tr: %d  \tg: %d \tb: %d\n",J,palette[J].r<<10,palette[J].g<<10,palette[J].b<<10);
#endif
    if(XAllocColor(Dsp,DefaultCMap,&Color))
    { I++;Pal[J]=Color.pixel; }
  }
  if(Verbose) printf("obtained %d colors\n",I);

#ifdef SOUND
  if(Verbose) printf("Init Sound\n");
  InitSound(UseSound,Verbose);
  SetChannels(SndVolume,SndSwitch);
  SetSound(3,SND_NOISE);
#endif SOUND

  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  printf("InitMachine END\n");
  return(1);

}

/** Joysticks ************************************************/
/** Check for keyboard events, parse them, and return the   **/
/** joystick controller status                              **/
/*************************************************************/
#define JOY_A   1
#define JOY_B   2
#define JOY_SELECT      4
#define JOY_START       8
#define JOY_UP  0x10
#define JOY_DOWN        0x20
#define JOY_LEFT        0x40
#define JOY_RIGHT       0x80

int Joysticks(void)
{
  static int  JS=0x0000;
  static byte N=0;
  XEvent E;
  word J;

  /* Clear PAUSE button flag */
  JS&=0xFFFF;

  if(XCheckWindowEvent(Dsp,Wnd,KeyPressMask|KeyReleaseMask,&E))
  {
    J=XLookupKeysym((XKeyEvent *)&E,0);
    if(E.type==KeyPress)
      switch(J)
      {
        case XK_Escape:
        case XK_F12:    //CPURunning=0; 
	JS|=0x10000;
	break;
#ifdef 0 //DEBUG
        case XK_F1:     Trace=!Trace;break;

#ifdef 0
        case XK_F2:
          printf("***** SprTab = %04Xh *****\n",SprTab-VRAM);
          for(J=0;J<64;J++)
            printf
            (
              "SPRITE#%02d: %d,%d  %02Xh  %02Xh\n",
              J,SprX(J),SprY(J),SprP(J),SprA(J)
            );
#endif
          break;
#endif
#ifdef SOUND
        case XK_1:      SetChannels(SndVolume,SndSwitch^=0x01);break;
        case XK_2:      SetChannels(SndVolume,SndSwitch^=0x02);break;
        case XK_3:      SetChannels(SndVolume,SndSwitch^=0x04);break;
        case XK_4:      SetChannels(SndVolume,SndSwitch^=0x08);break;
        case XK_0:      SndSwitch=SndSwitch? 0x00:0x0F;
                        SetChannels(SndVolume,SndSwitch);break;
        case XK_minus: case XK_underscore:
          if(SndVolume>0) SndVolume-=10;
          SetChannels(SndVolume,SndSwitch);
          break;
        case XK_plus: case XK_equal:
          if(SndVolume<250) SndVolume+=10;
          SetChannels(SndVolume,SndSwitch);
          break;
#endif
        case XK_z: case XK_x: case XK_c: case XK_v:
        case XK_b: case XK_n: case XK_m: 
        case XK_Z: case XK_X: case XK_C: case XK_V:
        case XK_B: case XK_N: case XK_M: 
        case XK_Control_L:
          JS|=JOY_B;break;

        case XK_a: case XK_s: case XK_d: case XK_f:
        case XK_g: case XK_h: case XK_j:
        case XK_A: case XK_S: case XK_D: case XK_F:
        case XK_G: case XK_H: case XK_J:
        case XK_space: case XK_Alt_L:
          JS|=JOY_A;break;

        case XK_Return: JS|=JOY_START;break;
        case XK_Tab:    JS|=JOY_SELECT;break;
        case XK_Down:   JS|=JOY_DOWN;break;
        case XK_Up:     JS|=JOY_UP;break;
        case XK_Left:   JS|=JOY_LEFT;break;
        case XK_Right:  JS|=JOY_RIGHT;break;
        case XK_BackSpace: JS|=0x10000;break;

        case XK_Caps_Lock: case XK_Shift_L: case XK_Shift_R: 
          JS=0x0000;N=1;break;
      }
    else
      switch(J)
      {
        case XK_z: case XK_x: case XK_c: case XK_v:
        case XK_b: case XK_n: case XK_m: 
        case XK_Z: case XK_X: case XK_C: case XK_V:
        case XK_B: case XK_N: case XK_M:
        case XK_Control_L:
          JS&=~JOY_B;break;

        case XK_a: case XK_s: case XK_d: case XK_f:
        case XK_g: case XK_h: case XK_j:
        case XK_A: case XK_S: case XK_D: case XK_F:
        case XK_G: case XK_H: case XK_J:
        case XK_space: case XK_Alt_L:
          JS&=~JOY_A;break;

        case XK_Return: JS&=~JOY_START;break;
        case XK_Tab:    JS&=~JOY_SELECT;break;
        case XK_Down:   JS&=~JOY_DOWN;break;
        case XK_Up:     JS&=~JOY_UP;break;
        case XK_Left:   JS&=~JOY_LEFT;break;
        case XK_Right:  JS&=~JOY_RIGHT;break;

        case XK_Caps_Lock: case XK_Shift_L: case XK_Shift_R:
          JS=0x0000;N=0;break;
      }
  }

  /* If saving CPU and focus is out, sleep */
  while(XCheckWindowEvent(Dsp,Wnd,FocusChangeMask,&E));
  if(SaveCPU&&(E.type==FocusOut))
  {
    XAutoRepeatOn(Dsp);   /* Turn autorepeat on       */
#ifdef SOUND
    StopSound();          /* Silence the sound        */
#endif
    while(!XCheckWindowEvent(Dsp,Wnd,FocusChangeMask,&E) /* &&CPURunning */ )
    {
      if(XCheckWindowEvent(Dsp,Wnd,ExposureMask,&E)) RefreshScreen();
      XPeekEvent(Dsp,&E);
    }
    XAutoRepeatOff(Dsp);  /* Turn autorepeat off      */
#ifdef SOUND
    ResumeSound();        /* Switch the sound back on */
#endif
  }

  /* Return the current joystick states */
  return(JS);
}


/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#ifdef 0
#include "Common.h"
#endif 

#endif /* UNIX */
